@@ -1,7 +1,7 @@
 // qjackctlGraph.cpp
 //
 /****************************************************************************
-   Copyright (C) 2003-2022, rncbc aka Rui Nuno Capela. All rights reserved.
+   Copyright (C) 2003-2023, rncbc aka Rui Nuno Capela. All rights reserved.
 
    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
@@ -50,6 +50,9 @@
 #include <QWheelEvent>
 #include <QKeyEvent>
 
+#include <QGestureEvent>
+#include <QPinchGesture>
+
 #include <algorithm>
 
 #include <cmath>
@@ -278,7 +281,17 @@ void qjackctlGraphPort::setPortTitle ( const QString& title )
 {
 	m_title = (title.isEmpty() ? m_name : title);
 
-	m_text->setPlainText(m_title);
+	static const int MAX_TITLE_LENGTH = 29;
+	static const QString ellipsis(3, '.');
+
+	QString text = m_title;
+	const int nlength = text.indexOf(':');
+	if (nlength >= 0)
+		text.remove(0, nlength + 1);
+	if (text.length() >= MAX_TITLE_LENGTH + ellipsis.length())
+		text = ellipsis + text.right(MAX_TITLE_LENGTH);
+
+	m_text->setPlainText(text);
 
 	QPainterPath path;
 	const QRectF& rect
@@ -1239,7 +1252,7 @@ static const char *ColorsGroup   = "/GraphColors";
 qjackctlGraphCanvas::qjackctlGraphCanvas ( QWidget *parent )
 	: QGraphicsView(parent), m_state(DragNone), m_item(nullptr),
 		m_connect(nullptr), m_rubberband(nullptr),
-		m_zoom(1.0), m_zoomrange(false),
+		m_zoom(1.0), m_zoomrange(false), m_gesture(false),
 		m_commands(nullptr), m_settings(nullptr),
 		m_selected_nodes(0), m_repel_overlapping_nodes(false),
 		m_edit_item(nullptr), m_editor(nullptr), m_edited(0),
@@ -1268,6 +1281,8 @@ qjackctlGraphCanvas::qjackctlGraphCanvas ( QWidget *parent )
 		SIGNAL(editingFinished()),
 		SLOT(editingFinished()));
 
+	QGraphicsView::grabGesture(Qt::PinchGesture);
+
 	m_editor->setEnabled(false);
 	m_editor->hide();
 }
@@ -1597,6 +1612,9 @@ void qjackctlGraphCanvas::connectPorts (
 // Mouse event handlers.
 void qjackctlGraphCanvas::mousePressEvent ( QMouseEvent *event )
 {
+	if (m_gesture)
+		return;
+
 	m_state = DragNone;
 	m_item = nullptr;
 	m_pos = QGraphicsView::mapToScene(event->pos());
@@ -1622,6 +1640,9 @@ void qjackctlGraphCanvas::mousePressEvent ( QMouseEvent *event )
 
 void qjackctlGraphCanvas::mouseMoveEvent ( QMouseEvent *event )
 {
+	if (m_gesture)
+		return;
+
 	int nchanged = 0;
 
 	QPointF pos = QGraphicsView::mapToScene(event->pos());
@@ -1662,10 +1683,8 @@ void qjackctlGraphCanvas::mouseMoveEvent ( QMouseEvent *event )
 						++nchanged;
 					}
 					// Original node position (for move command)...
-					QPointF pos1 = m_pos;
-					pos1.setX(4.0 * ::round(0.25 * pos1.x()));
-					pos1.setY(4.0 * ::round(0.25 * pos1.y()));
-					m_pos1 = pos1;
+					m_pos1 = m_pos;
+					snapPos(m_pos1);
 				}
 				else m_item = nullptr;
 			}
@@ -1675,19 +1694,12 @@ void qjackctlGraphCanvas::mouseMoveEvent ( QMouseEvent *event )
 				m_rubberband = new QRubberBand(QRubberBand::Rectangle, this);
 			}
 			// Set allowed auto-scroll margins/limits...
-			const QRect& rect = QGraphicsView::rect();
-			const qreal mx = 0.5f * rect.width();
-			const qreal my = 0.5f * rect.height();
-			m_rect1 = m_scene->itemsBoundingRect()
-				.marginsAdded(QMarginsF(mx, my, mx, my));
+			boundingRect(true);
 		}
 		break;
 	case DragMove:
 		// Allow auto-scroll only if within allowed margins/limits...
-		if (!m_rect1.contains(pos)) {
-			pos.setX(qBound(m_rect1.left(), pos.x(), m_rect1.right()));
-			pos.setY(qBound(m_rect1.top(),  pos.y(), m_rect1.bottom()));
-		}
+		boundingPos(pos);
 		QGraphicsView::ensureVisible(QRectF(pos, QSizeF(2, 2)), 8, 8);
 		// Move new connection line...
 		if (m_connect)
@@ -1739,8 +1751,7 @@ void qjackctlGraphCanvas::mouseMoveEvent ( QMouseEvent *event )
 		}
 		// Move current selected nodes...
 		if (m_item && m_item->type() == qjackctlGraphNode::Type) {
-			pos.setX(4.0 * ::round(0.25 * pos.x()));
-			pos.setY(4.0 * ::round(0.25 * pos.y()));
+			snapPos(pos);
 			const QPointF delta = (pos - m_pos);
 			foreach (QGraphicsItem *item, m_scene->selectedItems()) {
 				if (item->type() == qjackctlGraphNode::Type) {
@@ -1789,6 +1800,9 @@ void qjackctlGraphCanvas::mouseMoveEvent ( QMouseEvent *event )
 
 void qjackctlGraphCanvas::mouseReleaseEvent ( QMouseEvent *event )
 {
+	if (m_gesture)
+		return;
+
 	int nchanged = 0;
 
 	switch (m_state) {
@@ -2259,13 +2273,14 @@ bool qjackctlGraphCanvas::restoreNode ( qjackctlGraphNode *node )
 		return false;
 
 	m_settings->beginGroup(NodePosGroup);
-	const QPointF& node_pos
-		= m_settings->value('/' + nodeKey(node)).toPointF();
+	QPointF node_pos = m_settings->value('/' + nodeKey(node)).toPointF();
 	m_settings->endGroup();
 
 	if (node_pos.isNull())
 		return false;
 
+	boundingPos(node_pos);
+
 	node->setPos(node_pos);
 	return true;
 }
@@ -2305,13 +2320,12 @@ bool qjackctlGraphCanvas::restoreState (void)
 	m_settings->endGroup();
 
 	m_settings->beginGroup(CanvasGroup);
-	m_settings->setValue(CanvasRectKey, QGraphicsView::sceneRect());
 	const QRectF& rect = m_settings->value(CanvasRectKey).toRectF();
 	const qreal zoom = m_settings->value(CanvasZoomKey, 1.0).toReal();
 	m_settings->endGroup();
 
 	if (rect.isValid())
-		QGraphicsView::setSceneRect(rect);
+		m_rect1 = rect; // QGraphicsView::setSceneRect(m_rect1);
 
 	setZoom(zoom);
 
@@ -2337,7 +2351,7 @@ bool qjackctlGraphCanvas::saveState (void) const
 
 	m_settings->beginGroup(CanvasGroup);
 	m_settings->setValue(CanvasZoomKey, zoom());
-	m_settings->setValue(CanvasRectKey, QGraphicsView::sceneRect());
+	m_settings->setValue(CanvasRectKey, m_rect1);
 	m_settings->endGroup();
 
 	m_settings->beginGroup(ColorsGroup);
@@ -2711,4 +2725,91 @@ void qjackctlGraphCanvas::repelOverlappingNodesAll (
 }
 
 
+// Gesture event handlers.
+//
+bool qjackctlGraphCanvas::event ( QEvent *event )
+{
+	if (event->type() == QEvent::Gesture)
+		return gestureEvent(static_cast<QGestureEvent *> (event));
+	else
+		return QGraphicsView::event(event);
+}
+
+
+bool qjackctlGraphCanvas::gestureEvent ( QGestureEvent *event )
+{
+	if (QGesture *pinch = event->gesture(Qt::PinchGesture))
+		pinchGesture(static_cast<QPinchGesture *> (pinch));
+
+	return true;
+}
+
+
+void qjackctlGraphCanvas::pinchGesture ( QPinchGesture *pinch )
+{
+	switch (pinch->state()) {
+	case Qt::GestureStarted: {
+		const qreal scale_factor = zoom();
+		pinch->setScaleFactor(scale_factor);
+		pinch->setLastScaleFactor(scale_factor);
+		pinch->setTotalScaleFactor(scale_factor);
+		m_gesture = true;
+		break;
+	}
+	case Qt::GestureFinished:
+		m_gesture = false;
+		// Fall thru...
+	case Qt::GestureUpdated:
+		if (pinch->changeFlags() & QPinchGesture::ScaleFactorChanged)
+			setZoom(pinch->totalScaleFactor());
+		// Fall thru...
+	default:
+		break;
+	}
+}
+
+
+// Bounding margins/limits...
+//
+const QRectF& qjackctlGraphCanvas::boundingRect ( bool reset )
+{
+	if (!m_rect1.isValid() || reset) {
+		const QRect& rect = QGraphicsView::rect();
+		const qreal mx = 0.33 * rect.width();
+		const qreal my = 0.33 * rect.height();
+		m_rect1 = m_scene->itemsBoundingRect()
+			.marginsAdded(QMarginsF(mx, my, mx, my));
+	}
+
+	return m_rect1;
+}
+
+
+void qjackctlGraphCanvas::boundingPos ( QPointF& pos )
+{
+	const QRectF& rect = boundingRect();
+	if (!rect.contains(pos)) {
+		pos.setX(qBound(rect.left(), pos.x(), rect.right()));
+		pos.setY(qBound(rect.top(),  pos.y(), rect.bottom()));
+	}
+}
+
+
+// Snap into position helpers.
+//
+void qjackctlGraphCanvas::snapPos ( QPointF& pos ) const
+{
+	pos.setX(4.0 * ::round(0.25 * pos.x()));
+	pos.setY(4.0 * ::round(0.25 * pos.y()));
+}
+
+
+QPointF qjackctlGraphCanvas::snapPos ( qreal x, qreal y ) const
+{
+	QPointF pos(x, y);
+	snapPos(pos);
+	return pos;
+}
+
+
 // end of qjackctlGraph.cpp
